//****************************************************************************
//	File Name    : 'main.c'
//	Title        : interface to the ISRs
//	Author       : Lennart Schmidt - Copyright (C) 2010
//	Created      : 2010.01.26
//	Revised      :
//	Version      : 0.1
//	Target MCU   : Atmel ARM AT91SAM7S Series
//	Target Code  : ARM instruction set (32 bit)
//	Editor Tabs  : 4
//
//	NOTE: This code is currently below version 1.0, and therefore is
//		considered to be lacking in some functionality or documentation,
//		or may not be fully tested.
//		Nonetheless, you can expect most functions to work.
//****************************************************************************
//	This code is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 2 of the License, or
//	(at your option) any later version.
//
//	This code is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this code. If not, see <http://www.gnu.org/licenses/>.
//****************************************************************************


// AT91SAMT7S256 definitions
#include <AT91SAM7S256.h>
// Exact-width integer types. WG14/N843 C99 Standard
#include <stdint.h>
// interface to the ISRs (foreground)
#include "isr.h"
// ARM exceptions, including interrupt locking
#include "arm_exceptions.h"

#include "systimer.h"

// local event flags shared between foreground (ISRs) and the background .....
static uint32_t volatile l_eventFlags;    // uninitialized data (.bss section)


//............................................................................
void ISR_pit(void) {                  // Programmable Interval Timer (PIT) ISR
	// get number of intervals elapsed since the last irq
	uint32_t volatile picnt = AT91C_BASE_PITC->PITC_PIIR;
	picnt = ((picnt & AT91C_PITC_PICNT) >> 20);

	// update system counter
	systimer_add((systime_t)(1 + picnt));

	uint32_t volatile dummy = AT91C_BASE_PITC->PITC_PIVR;  // clear int source
	eventFlagSet(PIT_FLAG);                          // set the PIT event flag
	(void)dummy;            // suppress warning "dummy" was set but never used
}
//............................................................................
void ISR_timer0(void) {                                          // Timer0 ISR
    uint32_t volatile dummy = AT91C_BASE_TC0->TC_SR;       // clear int source
    eventFlagSet(TIMER0_FLAG);                    // set the TIMER0 event flag

    (void)dummy;            // suppress warning "dummy" was set but never used
}


//............................................................................
void eventFlagSet(uint8_t flag) {
    ARM_INT_KEY_TYPE int_lock_key;
    ARM_INT_LOCK(int_lock_key);
    l_eventFlags |= (1 << flag);
    ARM_INT_UNLOCK(int_lock_key);
}
//............................................................................
uint8_t eventFlagCheck(uint8_t flag) {
    register uint8_t isSet;
    ARM_INT_KEY_TYPE int_lock_key;
    ARM_INT_LOCK(int_lock_key);
    isSet = ((l_eventFlags & (1 << flag)) != 0);
    if (isSet) {
        l_eventFlags &= ~(1 << flag);                        // clear the flag
    }
    ARM_INT_UNLOCK(int_lock_key);
    return isSet;
}


//............................................................................
__attribute__ ((section (".text.fastcode")))
void ISR_irq(void) {                                                // IRQ ISR
	typedef void (*IntVector)(void);
    IntVector vect = (IntVector)AT91C_BASE_AIC->AIC_IVR;       // read the IVR
                                         // write IVR if AIC in protected mode
    AT91C_BASE_AIC->AIC_IVR = (AT91_REG)vect;

    asm("MSR cpsr_c,#(0x1F)");   // enable IRQ/FIQ to allow nesting interrupts
    (*vect)();             // call the IRQ handler via the pointer to function
    asm("MSR cpsr_c,#(0x1F | 0x80)");                // lock IRQ before return

                                         // write AIC_EOICR to clear interrupt
    AT91C_BASE_AIC->AIC_EOICR = (AT91_REG)vect;
}
//............................................................................
__attribute__ ((section (".text.fastcode")))
void ISR_fiq(void) {                                                // FIQ ISR
    // Handle the FIQ directly. No AIC vectoring overhead necessary
    uint32_t volatile dummy = AT91C_BASE_TC1->TC_SR;       // clear int source
    eventFlagSet(TIMER1_FLAG);               // for example, set an event flag

    (void)dummy;            // suppress warning "dummy" was set but never used
}
//............................................................................
void ISR_spur(void) {                                          // spurious ISR
}
//............................................................................
void ISR_abort(char const *msg) {
	/* this function is called when an exception occurs.
	 * For production code you need to log the message and go to fail-safe
	 * state. You might also want to reset the CPU.
	 */

	// for .. ever
	for (;;) { ; }
}
