// AT91SAMT7S256 definitions
#include <AT91SAM7S256.h>
// Exact-width integer types. WG14/N843 C99 Standard
#include <stdint.h>
// at91 library
#include "at91lib.h"
// drivers
#include "dogm128.h"
#include "dogm128_fonts.h"
// C header
#include <stdio.h>
#include <string.h>


#include "globals.h"
#include "isr.h"
#include "arm_exceptions.h"
#include "systimer.h"


#define BIT(n)	(1 << (n))
#define PIN_17	BIT(17)	// PA.17
#define PIN_18	BIT(18)	// PA.18
#define PIN_19	BIT(19)	// PA.19
#define PIN_20	BIT(20)	// PA.20
uint32_t const BSP_led[] = {                // constant data (.rodata section)
	PIN_17,		// yellow LED on Olimex SAM7-P256
	PIN_18		// green  LED on Olimex SAM7-P256
};
uint32_t const BSP_key[] = {                // constant data (.rodata section)
	PIN_20,		// button near the yellow LED on Olimex SAM7-P256
	PIN_19		// button near the green  LED on Olimex SAM7-P256
};
#define LED_ON(num_)	(AT91C_BASE_PIOA->PIO_CODR = BSP_led[num_])
#define LED_OFF(num_)	(AT91C_BASE_PIOA->PIO_SODR = BSP_led[num_])


void BSP_init(void);


//............................................................................
void BSP_init(void) {
	uint32_t i;

	/*
	 * basic hardware initialisation
	 */

	// enable PIO clock
	AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_PIOA);

	// initialize the LEDs...
	for (i = 0; i < sizeof(BSP_led)/sizeof(BSP_led[0]); ++i) {
		// enable pin
		AT91C_BASE_PIOA->PIO_PER = BSP_led[i];
		// configure as output pin
		AT91C_BASE_PIOA->PIO_OER = BSP_led[i];
		// extinguish the LED
		LED_OFF(i);
	}

	// initialize the KEYs...
	for (i = 0; i < sizeof(BSP_key)/sizeof(BSP_key[0]); ++i) {
		// enable pin
		AT91C_BASE_PIOA->PIO_PER = BSP_key[i];
		// enable input glitch filter
		AT91C_BASE_PIOA->PIO_IFER = BSP_key[i];
		// disable pull up resistor
		AT91C_BASE_PIOA->PIO_PPUDR = BSP_key[i];
	}

	/*
	 * ARM exception handlers
	 */

	// hook the exception handlers
	*(uint32_t volatile *)0x24 = (uint32_t)&ARM_undef;
	*(uint32_t volatile *)0x28 = (uint32_t)&ARM_swi;
	*(uint32_t volatile *)0x2C = (uint32_t)&ARM_pAbort;
	*(uint32_t volatile *)0x30 = (uint32_t)&ARM_dAbort;
	*(uint32_t volatile *)0x34 = (uint32_t)&ARM_reserved;
	*(uint32_t volatile *)0x38 = (uint32_t)&ARM_irq;
	*(uint32_t volatile *)0x3C = (uint32_t)&ARM_fiq;

	/*
	 * Advanced Interrupt Controller (AIC)
	 */

	// disable all interrupts
	AT91C_BASE_AIC->AIC_IDCR = ~0;
	// clear all interrupts
	AT91C_BASE_AIC->AIC_ICCR = ~0;
	// write AIC_EOICR 8 times
	for (i = 0; i < 8; ++i) {
		AT91C_BASE_AIC->AIC_EOICR = 0;
	}

	// configure the spurious interrupt
	AT91C_BASE_AIC->AIC_SPU = (AT91_REG)&ISR_spur;

	// configure the PIT interrupt (10.000 ticks per second / 10 ticks per millisecond)
	// it seems that a higher tick rate than 100.000ticks/s isn't really possible
	// best performance is given at around 10.000ticks/s
	i = ((F_CPU * 2) / 16 / 10000) - 1;
	AT91C_BASE_PITC->PITC_PIMR = (AT91C_PITC_PITEN | AT91C_PITC_PITIEN | i);
	AT91C_BASE_AIC->AIC_SVR[AT91C_ID_SYS] = (AT91_REG)&ISR_pit;
	AT91C_BASE_AIC->AIC_SMR[AT91C_ID_SYS] =
		(AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL | (AT91C_AIC_PRIOR_LOWEST));
	AT91C_BASE_AIC->AIC_ICCR = (1 << AT91C_ID_SYS);
	AT91C_BASE_AIC->AIC_IECR = (1 << AT91C_ID_SYS);
	// configure systimer
	systimer_setres(10000);

    // configure Timer0 interrupt as IRQ
/*
    AT91F_TC_open(AT91C_BASE_TC0,
                  (AT91C_TC_WAVE | AT91C_TC_WAVESEL_UP_AUTO | TC_CLKS_MCK2),
                  AT91C_ID_TC0);
    AT91C_BASE_AIC->AIC_IDCR = (1 << AT91C_ID_TC0);
    AT91C_BASE_AIC->AIC_SVR[AT91C_ID_TC0] = (AT91_REG)&ISR_timer0;
    AT91C_BASE_AIC->AIC_SMR[AT91C_ID_TC0] =
        (AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL | ISR_TIMER0_PRIO);
    AT91C_BASE_TC0->TC_IER = AT91C_TC_CPCS;                 //  IRQ enable CPC
    AT91C_BASE_AIC->AIC_ICCR = (1 << AT91C_ID_TC0);
    AT91C_BASE_AIC->AIC_IECR = (1 << AT91C_ID_TC0);
*/

    // configure Timer1 interrupt as FIQ
/*
    AT91F_TC_open(AT91C_BASE_TC1,
                  (AT91C_TC_WAVE | AT91C_TC_WAVESEL_UP_AUTO | TC_CLKS_MCK2),
                  AT91C_ID_TC1);
    AT91C_BASE_AIC->AIC_IDCR = (1 << AT91C_ID_TC1);
    AT91C_BASE_AIC->AIC_SMR[AT91C_ID_TC1] =
        (AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL | 0);     // ignore priority for FIQ
    AT91C_BASE_AIC->AIC_ICCR = (1 << AT91C_ID_TC1);
    AT91C_BASE_TC1->TC_IER = AT91C_TC_CPCS;                     //  enable CPC
    AT91C_BASE_AIC->AIC_FFER = (1 << AT91C_ID_TC1);      // Fast Forcing (FIQ)
    AT91C_BASE_AIC->AIC_IECR = (1 << AT91C_ID_TC1);
*/

/*
    AT91C_BASE_TC0->TC_CCR = AT91C_TC_SWTRG;                   // start Timer0
    AT91C_BASE_TC1->TC_CCR = AT91C_TC_SWTRG;                   // start Timer1
*/

    // unlock IRQ/FIQ at the ARM core level
	ARM_INT_UNLOCK(0x1F);
}
//............................................................................
void at91_uart0_send(char*text);
void at91_uart0_send(char*text) {
	int x = 0;
	while(text[x] != 0) {
		at91_uart0_putc(text[x]);
		x++;
	}
	at91_uart0_putc('\n');
	at91_uart0_putc('\r');
}
//............................................................................
char test[100];
int main(void) {

	// initialize hardware
	BSP_init();

	// initialize AT91 devices
	at91_spi_init();
	at91_uart0_init(UART_BAUD(115200), UART_8N1);

	// initialize drivers
	dogm128_init(DOGM128_SPI_DEVICE(0) // SPI device address
	           | DOGM128_SPI_CLKDIV(2) // SPI clock divisor
	           /* | DOGM128_REVERSE */);

	LED_ON(1);
	at91_uart0_send("Copyright 2010 - Lennart Schmidt");

	siprintf(test, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
//	siprintf(test, "GCC Version: %d.%d.%d", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);//"ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	dogm128_drawstring(0, 0, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, __DATE__);//"abcdefghijklmnopqrstuvwxyz");
	siprintf(test, "abcdefghijklmnopqrstuvwxyz");
	dogm128_drawstring(0, 8, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, "0123456789 ");
	dogm128_drawstring(0, 16, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, "@!\"$%%&/<({[|]})>?\\+*~#':;,.-_=");
	dogm128_drawstring(0, 24, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	dogm128_update();

/*
	siprintf(test, "0x18: 0x%08X", *(uint32_t volatile *)0x18);
	dogm128_drawstring(0, 0, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, "0x38: 0x%08X", *(uint32_t volatile *)0x38);
	dogm128_drawstring(0, 8, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, "ARM_irq: 0x%08X", &ARM_irq);
	dogm128_drawstring(0, 16, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);

	siprintf(test, "0x14: 0x%08X", *(uint32_t volatile *)0x14);
	dogm128_drawstring(0, 24, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	extern uint32_t __ram_start;
	siprintf(test, "__ram_start: %08X", &__ram_start);
	dogm128_drawstring(0, 32, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);

	dogm128_update();
*/

	uint32_t loop = 0;
	uint8_t blink = 1;

	uint16_t fps = 0;
	uint32_t c1 = (systimer_get() % 0x100000000);
	uint32_t c2 = 0;

/*
	siprintf(test, "Copyright 2010");
	dogm128_drawstring(0, 0, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
	siprintf(test, "Lennart Schmidt");
	dogm128_drawstring(0, 8, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
*/

	uint16_t mmmssWidth = dogm128_getstringwidth(FONT_LCD_24, "000:00");
	for(;;) { // for..ever

		systime_t pit = (systimer_get() % 0x100000000);

		uint32_t res = systimer_getres();
		int m = ((pit/res)/60)%1000;
		int s = (pit/res)%60;
		int ms = ((pit*1000)/res)%1000;
		siprintf(test, "%03d:%02d", m, s);
		dogm128_drawstring(0, 40, FONT_LCD_24, test, DOGM128_DRAW_OPAQUE);
		siprintf(test, ".%04d", ms);
		dogm128_drawstring(mmmssWidth, 48, FONT_LCD_16, test, DOGM128_DRAW_OPAQUE);

		c2 = pit;
		fps++;
		if((c2-c1) >= systimer_getres()) {
			siprintf(test, "FPS: %d  ", fps);
			dogm128_drawstring(mmmssWidth, 40, FONT_PROP_8, test, DOGM128_DRAW_OPAQUE);
			c1 = c2; fps = 0;
		}

		dogm128_update();

		loop++;
		if(loop > fps) {
			loop = 0;
			blink = (blink == 1) ? 0 : 1;
			if(blink == 1)
				LED_ON(0);
			else
				LED_OFF(0);
		}

	} // for .. ever

	// unreachable; suppress compiler warning
	return(0);
}
