//****************************************************************************
//	File Name    : 'low_level_init.c'
//	Title        : early initializations of the hardware
//	Author       : Lennart Schmidt - Copyright (C) 2010
//	Created      : 2010.01.26
//	Revised      :
//	Version      : 0.1
//	Target MCU   : Atmel ARM AT91SAM7S Series
//	Editor Tabs  : 4
//
//	NOTE: This code is currently below version 1.0, and therefore is
//		considered to be lacking in some functionality or documentation,
//		or may not be fully tested.
//		Nonetheless, you can expect most functions to work.
//****************************************************************************
//	This code is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 2 of the License, or
//	(at your option) any later version.
//
//	This code is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this code. If not, see <http://www.gnu.org/licenses/>.
//****************************************************************************


// AT91SAMT7S256 definitions
#include <AT91SAM7S256.h>
// Exact-width integer types. WG14/N843 C99 Standard
#include <stdint.h>

// global things
#include "globals.h"


/*****************************************************************************
 * low_level_init() is invoked by the startup sequence after initializing
 * the C stack, but before initializing the segments in RAM.
 *
 * low_level_init() is invoked in the ARM state. The function gives the
 * application a chance to perform early initializations of the hardware.
 * This function cannot rely on initialization of any static variables,
 * because these have not yet been initialized in RAM.
 */
void low_level_init(void (*reset_addr)(), void (*return_addr)()) {
	static uint32_t const LDR_PC_PC = 0xE59FF000U;
	static uint32_t const MAGIC = 0xDEADBEEFU;
	AT91PS_PMC pPMC;

	// Reset all peripherals
	AT91C_BASE_RSTC->RSTC_RCR = AT91C_RSTC_KEY | AT91C_RSTC_PERRST;
	// Enable User Reset and set its minimal assertion to 960 us
	AT91C_BASE_RSTC->RSTC_RMR = AT91C_RSTC_KEY | AT91C_RSTC_URSTEN;// | (0x4<<8);

	// The watchdog is enabled after processor reset. Disable it.
	AT91C_BASE_WDTC->WDTC_WDMR = AT91C_WDTC_WDDIS;

	// Set flash wait states and FMCN .. 48MHz = 1 wait state
	AT91C_BASE_MC->MC_FMR = ((AT91C_MC_FMCN) & ((F_CPU + 500000)/1000000 << 16))
						  | AT91C_MC_FWS_1FWS;

	/* Init PMC Step 1. Enable Main Oscillator
	 * Main Oscillator Startup Time worst case (3MHz) corresponds to 15ms
	 * (0x40 for AT91C_CKGR_OSCOUNT field)
	 */
	pPMC = AT91C_BASE_PMC;
	pPMC->PMC_MOR = ((0x40 << 8) & AT91C_CKGR_OSCOUNT) | AT91C_CKGR_MOSCEN;
	// Wait the startup time
	while ((pPMC->PMC_SR & AT91C_PMC_MOSCS) == 0) { ; }

	/* Init PMC Step 2.
	 *
	 * Set PLL to 96MHz (96,109MHz) and UDP Clock to 48MHz
	 * PLL Startup time depends on PLL RC filter: worst case is choosen
	 * UDP Clock (48,058MHz) is compliant with the Universal Serial Bus
	 * Specification (+/- 0.25% for full speed)
	 * - F_OSC = 18,432MHz
	 * - div by 14 Fin = 1,316MHz (18,432/14)
	 * - mul by 72+1 Fout = 96,109MHz (1,316*73)
	 */
	pPMC->PMC_PLLR = (AT91C_CKGR_USBDIV_1
				   | (AT91C_CKGR_DIV & 14)
				   | (AT91C_CKGR_PLLCOUNT & (16 << 8))
				   | (AT91C_CKGR_MUL & (72 << 16)));
	// Wait for PLL stabilization
	while ((pPMC->PMC_SR & AT91C_PMC_LOCK) == 0) { ; }
	// Wait until the Master Clock is established
	while ((pPMC->PMC_SR & AT91C_PMC_MCKRDY) == 0) { ; }

	/* Init PMC Step 3.
	 * Selection of Master Clock MCK equal to (Processor Clock PCK) PLL/2=48MHz
	 */
	pPMC->PMC_MCKR =  AT91C_PMC_PRES_CLK_2;
	// Wait until the Master Clock is established
	while ((pPMC->PMC_SR & AT91C_PMC_MCKRDY) == 0) { ; }

	pPMC->PMC_MCKR |= AT91C_PMC_CSS_PLL_CLK;
	// Wait until the Master Clock is established
	while ((pPMC->PMC_SR & AT91C_PMC_MCKRDY) == 0) { ; }

	/* Setup the exception vectors to RAM
	 *
	 * NOTE: the exception vectors must be in RAM *before* the remap
	 * in order to guarantee that the ARM core is provided with valid vectors
	 * during the remap operation.
	 */
	// setup the primary vector table in RAM
	*(uint32_t volatile *)(AT91C_ISRAM + 0x00) = LDR_PC_PC | 0x18; // reset
	*(uint32_t volatile *)(AT91C_ISRAM + 0x04) = LDR_PC_PC | 0x18; // undef
	*(uint32_t volatile *)(AT91C_ISRAM + 0x08) = LDR_PC_PC | 0x18; // swi
	*(uint32_t volatile *)(AT91C_ISRAM + 0x0C) = LDR_PC_PC | 0x18; // pAbort
	*(uint32_t volatile *)(AT91C_ISRAM + 0x10) = LDR_PC_PC | 0x18; // dAbort
	*(uint32_t volatile *)(AT91C_ISRAM + 0x14) = MAGIC;            // reserved
	*(uint32_t volatile *)(AT91C_ISRAM + 0x18) = LDR_PC_PC | 0x18; // irq
	*(uint32_t volatile *)(AT91C_ISRAM + 0x1C) = LDR_PC_PC | 0x18; // fiq

	// setup the secondary vector table in RAM
	*(uint32_t volatile *)(AT91C_ISRAM + 0x20) = (uint32_t)reset_addr;
	*(uint32_t volatile *)(AT91C_ISRAM + 0x24) = 0x04U; // undef
	*(uint32_t volatile *)(AT91C_ISRAM + 0x28) = 0x08U; // swi
	*(uint32_t volatile *)(AT91C_ISRAM + 0x2C) = 0x0CU; // pAbort
	*(uint32_t volatile *)(AT91C_ISRAM + 0x30) = 0x10U; // dAbort
	*(uint32_t volatile *)(AT91C_ISRAM + 0x34) = 0x14U; // reserved
	*(uint32_t volatile *)(AT91C_ISRAM + 0x38) = 0x18U; // irq
	*(uint32_t volatile *)(AT91C_ISRAM + 0x3C) = 0x1CU; // fiq

	// check if the Memory Controller has been remapped already
	if (MAGIC != (*(uint32_t volatile *)0x14)) {
		// perform Memory Controller remapping
		AT91C_BASE_MC->MC_RCR = 1;
	}
}
